/* CSD: Cornerstone disassembler

    Copyright (C) 2014  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

package info.seasip.csd;

import java.io.*;
import java.util.Vector;

/** Main class for the disassembler: Parses arguments, kicks everything off. 
 *
 * @author      John Elliott <jce@seasip.demon.co.uk>
 * @version     1.0
 * @since       2014-04-12 
 */
public class CSD
{
	// Used to read the .CTL file (which defines extra blocks)
	private Reader ctlReader;
	// Two files to read the .MME and .OBJ files.
	private RandomAccessFile mmeRfile, objRfile;
	// The MME header.
	private Header header;
	// The index of modules
	private ModuleIndex modIndex;
	// The index of globals
	private GlobalIndex globIndex;
	// The modules themselves
	private Module[] modules;
	// The modules' individual headers.
	ModuleHeader[] modheads;
	// The initial RAM image.
	Ram ram;
	// Generated symbols.
	SymbolTable symbols = new SymbolTable();
	// List of procedures on which to force manual disassembly.
	SymbolTable controls = new SymbolTable();

	/** Constructor.

	  @param mmes The .mme file to load.
	  @param objs The .obj file to load.
	  @param ctlr The .ctl file to load. Can be null if there is no 
			ctl file.
	*/
	public CSD(RandomAccessFile mmes, RandomAccessFile objs,
			Reader ctlr)
	{
		mmeRfile  = mmes;
		objRfile  = objs;
		ctlReader = ctlr;
	}


	/** Load the MME and OBJ files.

	@exception IOException If the load fails for any reason.
	*/
	public void load() throws IOException
	{
		int mc, n;

		header    = new Header(mmeRfile);
		modIndex  = new ModuleIndex(mmeRfile);
		globIndex = new GlobalIndex(mmeRfile);
		header.load();	
		modIndex.load();
		mc = modIndex.getModuleCount();
		globIndex.load(mc);

		mmeRfile.seek(header.getInitialRAMOffset());
		ram = new Ram(mmeRfile, header.getInitialRAMLen());
		ram.load();

		modheads = new ModuleHeader[mc];
		modules  = new Module[mc];
		mmeRfile.seek(header.getModuleHeaders());
		for (n = 0; n < mc; n++)
		{
			modheads[n] = new ModuleHeader(mmeRfile, n);
			modheads[n].load();
			modules[n] = new Module(n, objRfile, modheads[n], 
				globIndex.getModuleGlobals(n), this);
			modules[n].load(moduleAddress(n), moduleLength(n));
		}
		try
		{
			if (ctlReader != null)
			{
				controls.loadSymbols(ctlReader);
			}
		}
		catch (IOException ex)
		{
			System.err.println(ex);
		}

	}

	public void disasm(PrintStream out)
	{
		int startMod  = header.getStartModule() - 1;
		int startProc = header.getStartProc();
		int startOff  = modheads[startMod].getProcAddress(startProc);

		header.showContents(System.out);
		modIndex.showContents(System.out);		
		globIndex.showContents(System.out);
		for (int n = 0; n < modheads.length; n++)
			modheads[n].showContents(System.out);
		ram.showContents(System.out);

		/* Analyse */
		System.out.printf("; startMod  = %d\n",   startMod);	
		System.out.printf("; startProc = %d\n",   startProc);	
		System.out.printf("; startOff  = %04x\n", startOff);

		for (int n = 0; n < modules.length; n++)
		{
			modules[n].analyseModule(symbols);	
			/* Analyse user-specified areas */
			Symbol[] syms = controls.moduleSymbols(n + 1, 'P');
			for (Symbol s: syms)
			{
				System.err.printf("Manual analysis of %02d:%04x\n", s.module, s.offset);
				modules[n].analyseProc(s.offset, symbols);
			}
		}

		/* Print the results */
		Symbol sym = symbols.lookup(startMod, startOff);
		if (sym != null)
		{
			System.err.println("Start proc = " + sym.name);
			sym.name = "_start";
		}
		for (int n = 0; n < modheads.length; n++)
		{
			System.out.println("; ---------");
			System.out.printf ("; Module %d\n", n + 1);
			System.out.println("; ---------");
			modules[n].coverage(out, symbols);
			modules[n].disasm(out, symbols);
		}
		// modules[startMod].disasm(out, startOff);
	}

	int moduleAddress(int module)
	{
		if (module < modIndex.getModuleCount()) return modIndex.getModuleAddress(module);
		return header.getCodeEnd();
	}


	int moduleLength(int module)
	{
		return moduleAddress(module + 1) - moduleAddress(module);
	}


	public static void main(String args[])
	{
		CSD csd;
		File mmeFile, objFile, ctlFile;
		RandomAccessFile mmeRfile, objRfile;	
		Reader ctlReader;

		if (args.length < 1) 
			mmeFile = new File("corner.mme");
		else	mmeFile = new File(args[0]);

		String strArg = mmeFile.getPath();;
		int offs = strArg.indexOf(".mme");
		if (offs > -1) strArg = strArg.substring(0, offs);

		if (args.length < 2)
			objFile = new File(strArg + ".obj");
		else	objFile = new File(args[1]);

		if (args.length < 3)
			ctlFile = new File(strArg + ".ctl");
		else	ctlFile = new File(args[2]);
		try
		{
			ctlReader = new BufferedReader(new FileReader(ctlFile));
		}
		catch (FileNotFoundException e)
		{
			ctlReader = null;
		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.err.println("I/O exception: " + e.getMessage());
			ctlReader = null;
		}
		try
		{
			mmeRfile = new RandomAccessFile(mmeFile, "r");
			objRfile = new RandomAccessFile(objFile, "r");
			csd = new CSD(mmeRfile, objRfile, ctlReader);
			csd.load();
			csd.disasm(System.out);
		}
		catch (FileNotFoundException e)
		{
			System.err.println("File not found: " + e.getMessage());
			System.exit(-1);
		}
		catch (EOFException e)
		{
			e.printStackTrace();
			System.err.println("Unexpected end of file.");
			System.exit(-1);
		}
		catch (IOException e)
		{
			e.printStackTrace();
			System.err.println("I/O exception: " + e.getMessage());
			System.exit(-1);
		}
	}
}

/*
#include <stdio.h>

typedef unsigned char byte;
typedef unsigned short word;

static FILE *fpmme, *fpobj;
static char *fnmme = "corner.mme";
static char *fnobj = "corner.obj";

static word mmehead[48];
static word modcount;
static word moddir[26];
static word globc;
static word globm[26];
static word globdir[1024];
static word globmod[26][1024];
static word modheads[2048];

void readword(FILE *fp, word *w)
{
	byte b[2];

	if (fread(b, 1, 2, fp) < 2) 
	{
		perror("fread"); exit(1);
	}	
	*w = b[0] | (((word)b[1]) << 8);
}

void readwords(FILE *fp, word *w, int c)
{
	int n;
	for (n = 0; n < c; n++) readword(fp, &w[n]);	
}

void show_list(word *w, int count)
{
	int n, x;

	for (n = x = 0; n < count; n++)
	{
		printf("%04x", w[n]);
		if (x < 11) 
		{
			++x;
			putchar(' ');
		}
		else 
		{
			printf ("\n%-19.19s", "");
			x = 0;
		}

	}
}



void show_modheads(void)
{
	int m;
	word *w = modheads;
	word len;

	for (m = 0; m < modcount; m++)
	{
		printf("Module %02d header : ", m);
		len = *w;
		show_list(w, len + 1);
		w += len + 1;
		printf("\n");
	}
}


void show_headers(void)
{
	int m;

	printf("Module count     = %d\n", modcount);
	printf("Module directory = ");
	show_list(moddir, modcount);
	printf("\n");
	printf("Modules end      = %04x\n", moddir[modcount] = (mmehead[7] + 1) * 2);

	for (m = 0; m < modcount; m++)
		printf("Module %02d range: %04x00 - %04x00\n", 
			m, moddir[m], moddir[m+1]);

	printf("Globals count    = %d\n", globc);
	printf("Global values    : ");
	show_list(globdir, globc);
	printf("\n");	
	for (m = 0; m < modcount; m++)
	{
		printf("  M %02d glbls(%03d): ", m, globm[m]);
		show_list(globmod[m], globm[m]);
		printf("\n");
	}
}


int main(int argc, char **argv)
{
	int n;

	fpmme = fopen(fnmme, "rb");
	fpobj = fopen(fnobj, "rb");

	readwords(fpmme, mmehead, 48);	
	readword (fpmme, &modcount);
	readwords(fpmme, moddir, modcount);
	readword (fpmme, &globc);
	readwords(fpmme, globm, modcount);
	readwords(fpmme, globdir, globc);
	for (n = 0; n < modcount; n++)
	{
		readwords(fpmme, globmod[n], globm[n]);
	}
	show_headers();

	fseek(fpmme, 256L * mmehead[13], SEEK_SET);
	readwords(fpmme, modheads, mmehead[15]);
	show_modheads();

	printf("tell=%lx\n", ftell(fpmme));
	
	fclose(fpmme);
	fclose(fpobj);
	return 0;
}
*/
