/* CSD: Cornerstone disassembler

    Copyright (C) 2014  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

package info.seasip.csd;
import java.io.*;

/** Represents the header of the .MME file.
 *
 * @author      John Elliott <jce@seasip.demon.co.uk>
 * @version     1.0
 * @since       2014-04-12
 */
public class Header extends WordChunk
{
	/** Standard constructor.
	 * @param rf The .MME file.
	 */
	Header(RandomAccessFile rf)
	{
		super(rf);
	}

	/** Load the header.
	 * @throws IOException on read error. */
	public void load() throws IOException
	{
		read(48);	// Always 48 words.
	}


	/** Get the first file position not used by code. 
	 * @return An offset in the .OBJ file.
	 */
	public int getCodeEnd()
	{
		return (words[7] + 1) * 512;
	}

	/** Get the size of the initial RAM image.
	 * @return size in bytes.
	 */
	public int getInitialRAMLen()
	{
		return words[11] * 2;
	}

	/** Get the offset in the file of the initial RAM image.
	 * @return The offset in the .MME file.
	 */
	public int getInitialRAMOffset()
	{
		return 256 * (1 + words[13] + ((words[15] * 2) / 256));
	}

	/** Get the offset in the file of the module headers.
	 * @return The offset in the .MME file.
	 */
	public int getModuleHeaders()
	{
		return words[13] * 256;
	}


	/** Get the length of the module headers table.
	 * @return The length, in words.
	 */
	public int getModuleHeadersLen()
	{
		return words[15];
	}

	/** Get the module containing the start procedure.
	 * @return The module ID, 1-based.
	 */
	public int getStartModule()
	{
		return (words[8] >> 8) & 0xFF;	
	}

	/** Get the index of the start procedure.
	 * @return The procedure index.
	 */
	public int getStartProc()
	{
		return words[8] & 0xFF;
	}

	/** Dump the contents of the MME header.
	 * @param stream The stream to output to.
	 */
	public void showContents(PrintStream stream)
	{
		stream.println(";---------------------------------------");
		stream.println("; MME header:");
		stream.println(";---------------------------------------");
		for (int n = 0; n < 48; n++)
		{
			stream.printf("; %04x: %04x ",
				n*2, words[n] & 0xFFFF);
			switch (n)
			{
				case 7: stream.printf("; Last page with code: 0x%04x - 0x%04x", getCodeEnd() - 512, getCodeEnd() - 1); break;
				case 8: stream.printf("; Start address = Module 0x%02x Procedure 0x%02x", getStartModule(), getStartProc()); break;
				case 11: stream.printf("; Initial RAM image length %04x", getInitialRAMLen()); break;
				case 13: stream.printf("; Module headers at 0x%06x", getModuleHeaders()); break;
				case 15: stream.printf("; Module headers length=%04x", getModuleHeadersLen()); break;
			}
			stream.print("\n");
		}		
		stream.printf("; Initial RAM image at 0x%04x\n", getInitialRAMOffset());

	}

}

