/* CSD: Cornerstone disassembler

    Copyright (C) 2014  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

package info.seasip.csd;
import java.io.*;
import java.util.Vector;
import java.util.Deque;
import java.util.ArrayDeque;
import java.util.Collections;
import java.util.ArrayList;
import java.util.List;

/** Class corresponding to a module. Contains the bulk of the disassembly 
 * code.
 *
 * @author      John Elliott <jce@seasip.demon.co.uk>
 * @version     1.0
 * @since       2014-04-12
 */
class Module
{
	/** The file from which the bytecode is loaded. */
	protected RandomAccessFile file;

	/** The bytecode comprising this module. */
	protected byte bytecode[];

	/** Used to keep the code analyser from entering an infinite loop.
	 * For each byte in bytecode, a count of the number of times it has
	 * been analysed. Once a limit is hit, analysis stops. */ 
	protected byte hitMap[];

	/** This module's header. */
	private ModuleHeader header;
	/** This module's global variables. */
	private ModuleGlobals globals;
	/** The application. Used to access other modules as required. */
	private CSD parent;

	/** Emulated Cornerstone stack.
	 * During the second pass of disassembly, we do a limited simulation
	 * of execution -- the idea being that if a value is pushed and
	 * later used, we can possibly interpret what that value might be.
	 *
	 * stack holds the calculation stack for doing this. */ 
	private Deque<StackElement> stack = null;
	/** Emulated Cornerstone local variables for the current procedure. */
	private StackElement[] locals = null;

	/** An arbitrary fake 'file handle'. This is increased by one with
	 * each OPEN opcode, so that the various open file handles can 
	 * be distinguished. */
	private short fileHandle = 1;
	/** An arbitrary fake 'tuple ID'. This is incremented with each
	 * tuple allocation. */
	private short tupleID = 1;
	/** An arbitrary fake 'vector ID'. This is incremented with each
	 * vector allocation. */
	private short vectorID = 1;	/* Arbitrary fake 'vector ID' */

	/** The current column on screen.
	 *  0 =&gt; opcode;<br />
         *  1 =&gt; argument;<br />
         *  2 =&gt; comment.<br /> */
	private int column;

	/** The current module ID. Rather unfortunately, this is 0-based 
	 * rather than 1-based, so tends to get incremented wherever it 
	 * happens to be used. */
	private int moduleID;


	/** Constructor.
	 * @param id 0-based module ID.
	 * @param rf The OBJ file from which the bytecode will be loaded.
	 * @param mh The module's header from the MME file.
	 * @param mg The module globals from the MME file.
	 * @param p  The parent application. */
	Module(int id, RandomAccessFile rf, ModuleHeader mh, 
			ModuleGlobals mg, CSD p)
	{
		moduleID = id;
		file     = rf;
		header   = mh;
		globals  = mg;
		parent   = p;	
	}

	/** Load the bytecode.
	 * @param offset Offset of the bytecode in file.
	 * @param length Length of the bytecode.
	 *
	 * @exception IOException If the load fails for any reason. */
	void load(int offset, int length) throws IOException
	{
		bytecode = new byte[length];

		file.seek(offset);
		file.read(bytecode);
	}

	/** Disassemble this module.
	 *
	 * Call this after previously calling analyseModule(), so that
	 * symbols is correctly populated with the list of procedures.
	 *
	 * @param out The stream to which the disassembly will be written.
	 * @param symbols The symbols generated by analyseModule().
	 */
        public void disasm(PrintStream out, SymbolTable symbols)
	{
		int n;
		int offset = 0, pos = 0, max;
		List<Integer> procList = new ArrayList<Integer>();

		// Get the list of procedures in this module.
		Symbol sym[] = symbols.moduleSymbols(moduleID + 1, 'P');
		for (Symbol s : sym)
		{
			procList.add(s.offset);
		}
		// Sort the list by address.
		Collections.sort(procList);
		// For each procedure...
		for (n = 0; n < procList.size(); n++)
		{
			// Get its start address.
			offset = procList.get(n);
			Symbol s = symbols.lookup(moduleID + 1, offset);
			// If there's a gap between the previous procedure
			// and this one, dump it.
			if (offset > pos)
			{
				dumpData(out, pos, offset);
			}
			// If this is not the last procedure, ensure the
			// disassembly stops when the next procedure is 
			// encountered.
			if (n < procList.size() - 1)
			{
				max = procList.get(n + 1);	
			}
			// Otherwise stop at the end of the file.
			else	max = bytecode.length;

			// Disassemble the procedure.
			disasmProc(out, offset, symbols, max);
			pos = s.max;
		}
		// If there are data bytes after the last procedure, 
		// dump them.
		if (pos < bytecode.length)
		{
			dumpData(out, pos, bytecode.length);
		}
	}

	/** Realign the program counter to the next 256-byte block. 
	 *  Used to simulate the effect of the NEXTB instruction.
	 *  @param offset The address of the byte after the NEXTB instruction. 
	 *  @return The number of bytes to skip. */
	private int nextb(int offset)
	{
		return 0xFF - ((offset-1) & 0xFF);
	}

	/** Disassemble a procedure.
	 * @param out The output stream to which the disassembly should be sent
	 * @param offset The address of the procedure
	 * @param symbols The symbol table containing labels
	 * @param max The address of the next procedure, or bytecode.length.
	 *           Either way, a limit past which disassembly should not go.
	 */
	private void disasmProc(PrintStream out, int offset, 
		SymbolTable symbols, int max)
	{
		// Bail if a procedure is completely out of bounds.
		if (offset >= bytecode.length) 
		{
			Exception e = new Exception();
			e.printStackTrace();

			System.err.printf("Procedure %04x out of range for module %d\n", offset, moduleID + 1);
			return;
		}
		// Look up this procedure in the symbol table.
		int base = offset;
		int procHead = bytecode[offset] & 0xFF;
		Symbol sym = symbols.lookup(moduleID + 1, offset);
		int last = -1;

		// Procedure not found. That means it hasn't been analysed.
		if (sym == null || sym.max < 0)
		{
			System.err.printf("Analysis error: Proc %02X:%04x not studied\n", moduleID + 1, offset);
			return;
		}
		last = sym.max;
		if (last > max) last = max;

		// Create a simulated stack
		stack = new ArrayDeque<StackElement>();
		// And a simulated set of local variables.
		locals = new StackElement[procHead & 0x7F];

		// Initialise the locals to Unknown.
		for (int n = 0; n < locals.length; n++)
		{
			locals[n] = new StackElement(-1);
		}

		// Print the procedure header.
		if (out != null)
		{
			out.printf("%02d:%04x    [%02x]     ", 
			 	moduleID + 1, offset & 0xFFFF, procHead);
			if (sym != null) out.printf("%-10.10s", sym.name + ":");
			else		 out.printf("%-10.10s", "");
			out.printf("  ; Procedure header, %d local%s\n",
				procHead, ((procHead & 0x7F) == 1) ? "" : "s");
		}

		// If there are initial values for local variables, show 
		// them.
		if ((bytecode[offset] & 0x80) != 0)
		{
			byte bt = 0;
			short val;
			while ((bt & 0x80) == 0)
			{
				bt = bytecode[++offset];
				if (out != null) out.printf("%02d:%04x    [%02x ",
					moduleID + 1, offset, bt);
				// Byte value, expand to 16-bit word.
				if ((bt & 0x40) != 0) 
				{
					val = bytecode[++offset];
					if (out != null) out.printf("%02x]   ", val & 0xFF);
				}
				else	// Word value.
				{
					val = bytecode[offset+2];
					val = (short)((val << 8) | bytecode[offset+1]);
					if (out != null) out.printf("%04x] ", val & 0xFFFF);
					offset += 2;
				}
				if (out != null)
				{
					out.printf("; Local %d := %04x\n", bt & 0x3F, val);
				}
				setLocal(bt & 0x3F, new StackElement(val));
			}	
		}
		++offset;

		// Now we're past the procedure header, disassemble 
		// every instruction in the procedure.
		while (offset < last)
		{
			offset += disasmInst(out, offset, symbols);
		}
	}

	/** Look up the symbol associated with the word at addr.
	 * @param type The symbol type (P for procedure, L for jump target)
	 * @param addr The address in the bytecode of the word to look up.
	 * @return The symbol as a string. */
	private String wordSym(char type, int addr)
	{
		return offsetSym(type, wordFrom(addr));
	}


	/** Look up the symbol associated with a given word.
	 * @param type The symbol type (P for procedure, L for jump target)
	 * @param addr The word to look up.
	 * @return The symbol as a string. */
	private String offsetSym(char type, int offset)
	{
		String s;

		Symbol sym = parent.symbols.lookup(moduleID + 1, offset);

		if (sym != null) return sym.name;

		if (type == 'P' && offset >= 0 && offset < bytecode.length)
		{
// Causes too many false positives	procList.add(offset);
		}

		// Generate a generic symbol name.
		s = String.format("%c%02d%04X", type, moduleID + 1, offset);
		parent.symbols.add(type, s, moduleID + 1, offset);
		return s;

	}
	
	/** Look up the symbol associated with a far procedure reference.
	 * @param offset The address in the bytecode of the procedure
	 *              reference (first byte is procedure ID, second is
	 * 		module ID).
	 * @return The symbol as a string. */
	private String wordFar(int offset)
	{
		int proc = bytecode[offset] & 0xFF;
		int mod  = (bytecode[offset + 1] & 0xFF); /* 1-based module ID */

		// Out-of-range module number: don't add to the symbol table.
		if (mod <= 0 || mod > parent.modheads.length)
		{
			return String.format("?%d:P%d",
				mod, proc);
		}

		// Convert (module, procedure) to offset.
		int xoffset = parent.modheads[mod - 1].getProcAddress(proc);
	
		Symbol sym = parent.symbols.lookup(mod, xoffset);

		if (sym != null) return sym.name;

		String s = String.format("P%02d%04X", mod, xoffset);
		parent.symbols.add('P', s, mod, xoffset);

		return s;
	}

	/** Read a little-endian word from the bytecode.
	 * @param offset The address to read.
	 * @return The resulting word. */
	private int wordFrom(int offset)
	{
		int l = bytecode[offset] & 0xFF;
		int h = bytecode[offset + 1] & 0xFF;

		return (h << 8) | l;
	}


	/** Read a byte from the bytecode.
	 * @param offset The address to read.
	 * @return The resulting byte. */
	private int byteFrom(int offset)
	{
		byte l = bytecode[offset];

		return ( l & 0xFF );
	}

	/** Disassemble a jump instruction.
	 * @param out The stream to which disassembly should be written.
	 * @param cond The condition (if any) under which the jump is made.
	 * @param offset The offset of the byte following the JUMP opcode. 
	 * @param stacks The number of elements this opcode will take from
	 *               the stack.
	 * @return The number of bytes following the JUMP opcode. */
	private int disasmJump(PrintStream out, String cond, int offset,	
				int stacks)
	{
		String s;
		int target;
		StackElement[] se = null;

		s = "JUMP" + cond;
		printOpcode(out, s);

		if (stacks > 0) 
		{
			se = new StackElement[stacks];
			for (int n = 0; n < se.length; n++)
			{
				se[n] = pop();
			}
		}
		if (bytecode[offset] == 0)
		{
			s = wordSym('L', offset + 1);
			printArg(out, s);	
			if (stacks > 0) printStack(out, se);
			newline(out);
			return 3;	
		}
		else
		{
			int b = bytecode[offset];
	
			if (b < 0) target = offset + 1 + b;
			else	   target = offset + 1 + b;

			printArg(out, offsetSym('L', target));

			if (b < 0) s = "-" + Integer.toString(-b);
			else	   s = "+" + Integer.toString(b);
			if (stacks > 0)
			{
				s += "  ";
				for (int n = 0; n < se.length; n++)
					s += String.format(" (%s)", 
						printSE(se[n]));
			}
			printStack(out, s);
			newline(out);	
			return 1;
		} 
	}

	/** Disassemble an orphaned jump instruction. Used for bits of code
	 * that look like JUMPs but weren't picked up by the code analyser
	 * and so are probably not reachable.
 	 *
	 * @param out The stream to which disassembly should be written.
	 * @param offset The offset of the byte following the JUMP opcode. 
	 */
	private int guessJump(PrintStream out, int offset)
	{
		String s;
		int target;
		StackElement se = null;

		// This is largely copied and pasted from disasmJump() above,
		// except that the jump is never conditional.
		s = "jump";
		printOpcode(out, s);

		if (bytecode[offset] == 0)
		{
			printArg(out, String.format("%04X", wordFrom(offset+1)));	
			newline(out);
			return 3;	
		}
		else
		{
			int b = bytecode[offset];
	
			if (b < 0) target = offset + 1 + b;
			else	   target = offset + 1 + b;

			printArg(out, String.format("%04X", target));	

			if (b < 0) s = "-" + Integer.toString(-b);
			else	   s = "+" + Integer.toString(b);
			printStack(out, s);
			newline(out);	
			return 1;
		} 
	}

	/** Print the contents of a StackElement.
	 * @param se The element to print.
	 * @return Its printable representation. */
	private String printSE(StackElement se)
	{
		// Value -1 displays as 'unknown'; others as numbers.
		if      (se.value == -1) return String.format("%s:????",
							se.type);
		else return String.format("%s:%04x", se.type, se.value);
	}

	/** Interpret the contents of a StackElement as an address in VM RAM,
	 *  and display the string at that address.
	 * 
	 * @param se The StackElement to display.
	 * @return The string value. 
	 */
	private String printSEString(StackElement se)
	{
		int len, offset;

		// If se value is unknown, display as printSE() would.
		if      (se.value == -1) return String.format("%s:????",
							se.type);

		// Otherwise treat as a Pascal-style string.
		StringBuffer s = new StringBuffer("\"");
		offset = se.value * 2;
		len = parent.ram.peekw(offset);

		offset += 2;
		while (len > 0)
		{
			char c = ((char)parent.ram.peek(offset++));
			// If string contains any not-printable characters,
			// bail.
			if (c < 0x20 || c > 0x7E)
				return String.format("%s:????", se.type);
			s.append(c);
			--len;
		}
		s.append("\"");
		return s.toString();
	}

	/** Pop an element from the simulated stack.
	 * Because the simulation is far from perfect, the bytecode may 
	 * try to pop an element from the stack when the stack is empty. 
	 * Cope with that by returning an indeterminate StackElement.
	 *
	 * @return The top element of the stack.
	 */
	private StackElement pop()
	{
		if (stack.isEmpty()) return new StackElement(-1);
		return stack.removeFirst();
	}

	/** Read a local variable.
	 * Because the simulation is far from perfect, don't treat an
	 * out-of-range local variable as an error. 
	 * @param n The number of the local variable.
	 * @return The value of the variable.
	 */
	private StackElement getLocal(int n)
	{
		if (n >= 0 && n < locals.length) return locals[n];

		return new StackElement(-1, "!ovflw");
	}

	/** Write a local variable.
	 * Because the simulation is far from perfect, don't treat an
	 * out-of-range local variable as an error -- just ignore it.
	 *
	 * @param n The number of the local variable.
	 * @param se The value to assign to it.
	 */
	private void setLocal(int n, StackElement se)
	{
		if (n >= 0 && n < locals.length) locals[n] = se;
	}

	/** Push a number onto the simulated stack.
	 * @param n The number to push. It will be given the description
	 * 'int'. -1 will be pushed as 0xFFFF, not -1. */
	private void push(int n)
	{
		stack.addFirst(new StackElement(n & 0xFFFF));
	}
	
	/** Push a number onto the simulated stack.
	 * @param n The number to push.
	 * @param t Its description. */
	private void push(int n, String t)
	{
		stack.addFirst(new StackElement(n, t));
	}


	/** Print an entry in the opcode column.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param opcode The opcode to print.	
	 */
	private void printOpcode(PrintStream out, String opcode)
	{
		if (out != null) out.printf("%-10.10s  ", opcode);
		column = 1;	
	}

	/** Move to next line of the disassembly. Resets the column count.
	 *
	 * @param out The stream to which disassembly is being written.
	 */
	private void newline(PrintStream out)
	{
		if (out != null) out.print('\n');
		column = 0;	
	}
	
	/** Print an entry in the argument column.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param args The argument(s) to print.
	 */
	private void printArg(PrintStream out, String args)
	{
		if (out != null) out.printf("%-10.10s  ", args);
		column = 2;	
	}
	
	/** Print an entry in the argument column.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param arg The argument to print, 0-255.
	 */
	private void printArg(PrintStream out, int arg)
	{
		if (out != null) out.printf("%-10.10s  ", 
			String.format("%d", arg & 0xFF));
		column = 2;	
	}

	/** Print an entry in the stack column. Inserts a blank argument 
	 * column if necessary.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param args The string to print.
	 */
	private void printStack(PrintStream out, String args)
	{
		if (column == 1) printArg(out, "");
		if (out != null) out.printf("; %s", args);
	}

	/** Print an entry in the stack column. Inserts a blank argument 
	 * column if necessary.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param arg1 The stack element to print.
	 */
	private void printStack(PrintStream out, StackElement arg1)
	{
		if (column == 1) printArg(out, "");
		if (out != null) out.printf("; (%s)", printSE(arg1));
	}

	/** Print an entry in the stack column. Inserts a blank argument 
	 * column if necessary.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param arg1 The first stack element to print.
	 * @param arg2 The second stack element to print.
	 */
	private void printStack(PrintStream out, StackElement arg1, StackElement arg2)
	{
		if (column == 1) printArg(out, "");
		if (out != null) out.printf("; (%s,%s)", printSE(arg1), printSE(arg2));
	}

	/** Print an entry in the stack column. Inserts a blank argument 
	 * column if necessary.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param arg1 The first stack element to print.
	 * @param arg2 The second stack element to print.
	 * @param arg3 The third stack element to print.
	 */
	private void printStack(PrintStream out, StackElement arg1, StackElement arg2, StackElement arg3)
	{
		if (column == 1) printArg(out, "");
		if (out != null) out.printf("; (%s,%s,%s)", printSE(arg1), printSE(arg2), printSE(arg3));
	}
	
	/** Print an entry in the stack column. Inserts a blank argument 
	 * column if necessary.
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param args An array of stack elements to print.
	 */
	private void printStack(PrintStream out, StackElement[] args)
	{
		if (column == 1) printArg(out, "");
		if (out != null)
		{
			out.printf("; (");
			for (int n = 0; n < args.length; n++)
			{
				if (n > 0) out.print(",");
				out.print(printSE(args[n]));
			}
		}
	}


	/** Print an extra row with an entry only in the stack column. 
	 *
	 * @param out The stream to which disassembly is being written.
	 * @param arg A stack element specifying how many further 
	 *           stack elements should be popped off the stack and 
	 *           printed.
	 */
	private void printStackChain(PrintStream out, StackElement arg)
	{
		StackElement se;

		if (arg.value != -1)
		{
			StringBuffer val = new StringBuffer();

			newline(out);
			if (out != null) out.printf("%-27s", "");
			printOpcode(out, "");
			val.append(" {");
			for (int n = 0; n < arg.value; n++)
			{
				se = pop();
				val.append(printSE(se));
				val.append(" ");
			}
			val.append("}");
			printStack(out, val.toString());	
		}
	}


	/** Disassemble an opcode. 
	 *  @param out The stream to which disassembly is being written.
	 *  @param offset The offset of the opcode to disassemble.
	 *  @param symbols The results of the analysis pass.
	 */
	private int disasmInst(PrintStream out, int offset, SymbolTable symbols)
	{
		StackElement se1, se2, se3;
		StackElement[] se;
		int l = 1;
		int i = bytecode[offset] & 0xFF;
		boolean nextLine = false;

		// Is there a symbol for this line?
		Symbol sym = symbols.lookup(moduleID + 1, offset);

		if (out != null)
		{
			// Print opcode byte(s). 0x5F is the first byte
			// of a 2-byte opcode.
			if (i == 0x5F)
			{
				out.printf("%02d:%04x    [%02x %02x]  ", 
					moduleID + 1, offset & 0xFFFF, i, 
					bytecode[offset+1] & 0xFF);
			}
			else	
			{
				out.printf("%02d:%04x    [%02x]     ", 
					moduleID + 1, offset & 0xFFFF, i);
			}
			// If there is a label here, print it.
			if (sym != null) out.printf("%-10.10s", sym.name + ":");
			else		 out.printf("%-10.10s", "");
		}
		++offset;
		column = 0;
		// Display the opcode.
		switch(i)
		{
			case 0x00:	printOpcode(out, "BREAK");
					newline(out);
					break;

			case 0x53:
			case 0x01:	se1 = pop();
					se2 = pop();
					printOpcode(out, "ADD");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se1.value + se2.value);
					break;
			case 0x02:	se1 = pop();
					se2 = pop();
					printOpcode(out, "SUB");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se2.value - se1.value);
					break;
			case 0x03:	se1 = pop();
					se2 = pop();
					printOpcode(out, "MUL");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se2.value * se1.value);
					break;
			case 0x04:	se1 = pop();
					se2 = pop();
					printOpcode(out, "DIV");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1
						|| se2.value == 0)
						push(-1, "unknown");
					else	push(se1.value / se2.value);
					break;
			case 0x05:	se1 = pop();
					se2 = pop();
					printOpcode(out, "MOD");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1
						|| se2.value == 0)
						push(-1, "unknown");
					else	push(se1.value % se2.value);
					break;
			case 0x06:	se1 = pop();
					printOpcode(out, "NEG");
					printStack  (out, se1);
					newline(out);
					if (se1.value == -1)
						push(-1, "unknown");
					else	push(-se1.value);
					break;
	
			case 0x07:	se1 = pop();
					se2 = pop();

					printOpcode(out, "ASHIFT");
					printStack(out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else
					{
						short sh = (short)(se1.value);
						short v =  (short)(se2.value);
					
						if (sh < 0) push(v >> (-sh));	
						else        push(v << sh);
					}
					break;
			case 0x08:	printOpcode(out, "INCL");
					printArg(out, byteFrom(offset));
					newline(out);
					se1 = getLocal(byteFrom(offset));
					if (se1.value == -1) 
						push(-1, "unknown");
					else 	push(se1.value + 1, se1.type);
				
					l++;
					break;
			case 0x09:	printOpcode(out, "PUSH_8");
					newline(out);
					push(8);
					break;
			case 0x0A:	printOpcode(out, "PUSH_4");
					newline(out);
					push(4);
					break;
			case 0x0B:	printOpcode(out, "DECL");
					printArg(out, byteFrom(offset));
					newline(out);
					se1 = getLocal(byteFrom(offset));
					if (se1.value == -1) 
						push(-1, "unknown");
					else 	push((se1.value - 1) & 0xFFFF, se1.type);
				
					l++;
					break;
			case 0x0C:	printOpcode(out, "PUSH_m1");
					newline(out);
					push(0xFFFF);
					break;
			case 0x0D:	printOpcode(out, "PUSH_3");
					newline(out);
					push(3);
					break;

			case 0x0E:	se1 = pop();
					se2 = pop();
					printOpcode(out, "AND");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se1.value & se2.value);
					break;
			case 0x0F:	se1 = pop();
					se2 = pop();
					printOpcode(out, "OR");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se1.value | se2.value);
					break;
			case 0x10:	se1 = pop();
					se2 = pop();

					printOpcode(out, "SHIFT");
					printStack(out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else
					{
						short sh = (short)(se1.value);
						int   v =  se2.value & 0xFFFF;
					
						if (sh < 0) push(v >> (-sh));	
						else        push(v << sh);
					}
					break;
			case 0x11:	se1 = pop();
					printOpcode(out, "VALLOC");
					printStack(out, se1);
					newline(out);
					push(vectorID++, "vector");
					break;
			case 0x12:	se1 = pop();
					printOpcode(out, "VALLOCI");
					printStack(out, se1);
					printStackChain(out, se1);
					newline(out);
					push(vectorID++, "vector");
					break;
			case 0x13:	se1 = pop();
					se2 = pop();
					printOpcode(out, "V?SIZE");
					printStack(out, se1, se2);
					newline(out);
					break;
			case 0x14:	se1 = pop();
					printOpcode(out, "TALLOC");
					printStack(out, se1);
					newline(out);
					push(tupleID++, "tuple");
					break;
			case 0x15:	se1 = pop();
					printOpcode(out, "TALLOCI");
					printStack(out, se1);
					printStackChain(out, se1);
					newline(out);
					push(tupleID++, "tuple");
					break;
			case 0x16:	se1 = pop();
					se2 = pop();
					printOpcode(out, "VLOADW");
					printStack(out, se1, se2);
					newline(out);
					push(-1, "unknown");
					break;
			case 0x17:	se1 = pop();
					se2 = pop();
					printOpcode(out, "VLOADB");
					printStack(out, se1, se2);
					newline(out);
					push(-1, "unknown");
					break;
			case 0x18:	se1 = pop();
					printOpcode(out, "VLOADW_");
					printArg(out, byteFrom(offset));
					printStack(out, se1);
					newline(out);
					push(-1, "unknown");
					++l;
					break;
			case 0x19:	se1 = pop();
					printOpcode(out, "VLOADB_");
					printArg(out, byteFrom(offset));
					printStack(out, se1);
					newline(out);
					push(-1, "unknown");
					++l;
					break;
			case 0x1A:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VPUTW");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x1B:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VPUTB");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x1C:	se1 = pop();
					se2 = pop();
					printOpcode(out, "VPUTW_");
					printArg(out, byteFrom(offset));
					++l;
					printStack(out, se1, se2);
					newline(out);
					break;
			case 0x1D:	se1 = pop();
					se2 = pop();
					printOpcode(out, "VPUTB_");
					printArg(out, byteFrom(offset));
					++l;
					printStack(out, se1, se2);
					newline(out);
					break;
			case 0x1E:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VECSETW");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x1F:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VECSETB");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x20:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VECCPYW");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x21:	se = new StackElement[5];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					se[4] = pop();
					printOpcode(out, "VECCPYB");
					printStack(out, se);
					newline(out);
					break;
			case 0x22:	printOpcode(out, "LOADG");
					printArg(out, byteFrom(offset));
					newline(out); 
					++l;    
					push(byteFrom(offset), "global");
					break;
			case 0x23:	printOpcode(out, "LOADMG");
					printArg(out, byteFrom(offset));
					newline(out); 
					++l;    
					push(byteFrom(offset), "mg");
					break;
			case 0x24:	printOpcode(out, "PUSH_2");
					newline(out);
					push(2);
					break;
			case 0x25:	printOpcode(out, "PUSHW");
					printArg(out, String.format("%04x",
						wordFrom(offset))); 
					newline(out);
					push(wordFrom(offset));
					l += 2; 
					break;
			case 0x26:	printOpcode(out, "PUSHB");
					printArg(out, String.format("%04x",
						byteFrom(offset))); 
					newline(out);
					push(byteFrom(offset));
					l += 1;
					break; 
			case 0x27:	printOpcode(out, "PUSH_NIL");	
					newline(out);
					push(0x8001, "nil");
					break;
			case 0x28:	printOpcode(out, "PUSH_0");	
					newline(out);
					push(0);
					break;
			case 0x29:	printOpcode(out, "DUP");
					newline(out);
					se1 = pop();
					stack.addFirst(se1);
					stack.addFirst(se1);
					break;
			case 0x2A:	printOpcode(out, "PUSH_m8");	
					push(0xFFF8);
					newline(out);
					break;
			case 0x2B:	printOpcode(out, "PUSH_5");	
					newline(out);
					push(5);
					break;
			case 0x2C:	printOpcode(out, "PUSH_1");	
					newline(out);
					push(1);
					break;
			case 0x2D:	se1 = pop();
					printOpcode(out, "PUTMG");
					printArg(out, byteFrom(offset));
					printStack(out, se1);
					newline(out);
					l++;	
					break;
			case 0x2E:	printOpcode(out, "PUSH_255");	
					push(255);
					newline(out);
					break;
			case 0x2F:	se1 = pop();
                                        printOpcode(out, "PULL");
					printStack(out, se1);
					newline(out);
					break;
			case 0x30:	l += disasmJump(out, "", offset, 0); 
					nextLine = true;
					break;
			case 0x31:	l += disasmJump(out, "Z", offset, 1); break;
			case 0x32:	l += disasmJump(out, "NZ", offset, 1); break;
			case 0x33:	l += disasmJump(out, "F", offset, 1); break;
			case 0x34:	l += disasmJump(out, "NF", offset, 1); break;
			case 0x35:	l += disasmJump(out, "G", offset, 1); break;
			case 0x36:	l += disasmJump(out, "LE", offset, 1); break;
			case 0x37:	l += disasmJump(out, "L", offset, 1); break;
			case 0x38:	l += disasmJump(out, "GE", offset, 1); break;
			case 0x39:	l += disasmJump(out, "L2", offset, 1); break;
			case 0x3A:	l += disasmJump(out, "LE2", offset, 1); break;
			case 0x3B:	l += disasmJump(out, "GE2", offset, 2); break;
			case 0x3C:	l += disasmJump(out, "G2", offset, 2); break;
			case 0x3D:	l += disasmJump(out, "EQ", offset, 2); break;
			case 0x3E:	l += disasmJump(out, "NE", offset, 2); break;
			case 0x3F:	printOpcode(out, "CALL0");
					printArg(out, wordSym('P', offset));
					newline(out);
					l += 2;
					break;
			case 0x40:	se1 = pop();
					printOpcode(out, "CALL1");
					printArg(out, wordSym('P', offset));
					printStack(out, se1);
					newline(out);
					l += 2;
					break;
			case 0x41:	se1 = pop();
					se2 = pop();
					printOpcode(out, "CALL2");
					printArg(out, wordSym('P', offset));
					printStack(out, se1, se2);
					newline(out);
					l += 2;
					break;
			case 0x42:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "CALL3");
					printArg(out, wordSym('P', offset));
					printStack(out, se1, se2, se3);
					newline(out);
					l += 2;
					break;
			case 0x43:	se1 = pop();
					printOpcode(out, "CALL");
					printArg(out, String.format("%d",
						byteFrom(offset)));
					se = new StackElement[byteFrom(offset) + 1];
					for (int n = 0; n < se.length; n++)
					{
						se[n] = pop();
					}
					printStack(out, se);
					newline(out);
					l++;
					break;

			case 0x44:	printOpcode(out, "CALLF0");
					printArg(out, wordFar(offset));
					newline(out);
					l += 2;
					break;

			case 0x45:	se1 = pop();
					printOpcode(out, "CALLF1");
					printArg(out, wordFar(offset));
					printStack(out, se1);
					newline(out);
					l += 2;
					break;

			case 0x46:	se1 = pop();
					se2 = pop();
					printOpcode(out, "CALLF2");
					printArg(out, wordFar(offset));
					printStack(out, se1, se2);
					newline(out);
					l += 2;
					break;
			case 0x47:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "CALLF3");
					printArg(out, wordFar(offset));
					printStack(out, se1, se2);
					newline(out);
					l += 2;
					break;
			case 0x48:	printOpcode(out, "CALLF");
					printArg(out, String.format("%d",
						byteFrom(offset)));
					se = new StackElement[byteFrom(offset) + 1];
					for (int n = 0; n < se.length; n++)
					{
						se[n] = pop();
					}
					printStack(out, se);
					newline(out);
					l++;
					break;

			case 0x49:	se1 = pop();
					printOpcode(out, "RETURN");
					printStack(out, se1);
					newline(out);
					l = 1;
					nextLine = true;
					break;
			case 0x4A:	printOpcode(out, "RFALSE");
					newline(out);
					l = 1;
					nextLine = true;
					break;
			case 0x4B:	printOpcode(out, "RZERO");
					newline(out);
					l = 1;
					nextLine = true;
					break;
			case 0x4C:	printOpcode(out, "PUSH_6");	
					newline(out);
					push(6);
					break;
			case 0x4D:	printOpcode(out, "HALT");
					printArg(out, String.format("%d",
						wordFrom(offset)));
					newline(out);
					nextLine = true;
					l += 2;
					break;
			case 0x4E:	printOpcode(out, "NEXTB");	
					newline(out);
					l += nextb(offset); 
					break;
			case 0x4F:	printOpcode(out, "PUSH_7");	
					newline(out);
					push(7);
					break;
			case 0x50:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "PRINTV");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x51:	se1 = pop();
					se2 = pop();
					printOpcode(out, "LOADVB2");
					printStack(out, se1, se2);
					newline(out);
					push(-1, "unknown");
					break;
			case 0x52:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "PUTVB2");
					printStack(out, se1, se2, se3);
					newline(out);
					break;
			case 0x54:	printOpcode(out, "INCLV");
					printArg(out, byteFrom(offset));
					newline(out);
					se1 = getLocal(byteFrom(offset));
					if (se1.value == -1) 
						push(-1, "unknown");
					else 	push(se1.value + 1, se1.type);
				
					l++;
					break;
			case 0x55:	printOpcode(out, "RET");
					newline(out);
					nextLine = true;
					break;
			case 0x56:	se1 = pop();
					printOpcode(out, "PUTL");
					printArg(out, byteFrom(offset));
					setLocal(byteFrom(offset), se1);
					printStack(out, se1);
					newline(out);
					++l;
					break;
			case 0x57:	printOpcode(out, "LOADL");
					printArg(out, byteFrom(offset));
					se1 = getLocal(byteFrom(offset));
					printStack(out, se1);
					push(se1.value, se1.type);
					newline(out);
					++l;
					break;
			case 0x58:	se1 = pop();
					push(se1.value, se1.type);
					printOpcode(out, "STOREL");
					printArg(out, byteFrom(offset));
					setLocal(byteFrom(offset), se1);
					printStack(out, se1);
					newline(out);
					++l;
					break;
			case 0x59:	printOpcode(out, "BITSVL");
					printArg(out, String.format("%d,%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 4) & 0x0F,
					(wordFrom(offset) >> 0) & 0x0F));
					newline(out);
					l += 2;
					push(-1, "unknown");
					break;
			case 0x5A:	se1 = pop();
					printOpcode(out, "BITSV");
					printArg(out, String.format("%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 0) & 0xFF));
					printStack(out, se1);
					newline(out);
					l += 2;
					push(-1, "unknown");
					break;
			case 0x5B:	se1 = pop();
					printOpcode(out, "BBSETVL");
					printArg(out, String.format("%d,%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 4) & 0x0F,
					(wordFrom(offset) >> 0) & 0x0F));
					printStack(out, se1);
					newline(out);
					l += 2;
					break;

			case 0x5C:	se1 = pop();
					se2 = pop();
					printOpcode(out, "BBSETV");
					printArg(out, String.format("%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 0) & 0xFF));
					printStack(out, se1, se2);
					newline(out);
					l += 2;
					break;
			case 0x5D:	se1 = pop();
					printOpcode(out, "BSETVL");
					printArg(out, String.format("%d,%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 4) & 0x0F,
					(wordFrom(offset) >> 0) & 0x0F));
					printStack(out, se1);
					newline(out);
					l += 2;
					break;

			case 0x5E:	se1 = pop();
					se2 = pop();
					printOpcode(out, "BSETV");
					printArg(out, String.format("%d,%d,%d", 
					(wordFrom(offset) >> 12) & 0x0F,
					(wordFrom(offset) >> 8) & 0x0F,
					(wordFrom(offset) >> 0) & 0xFF));
					printStack(out, se1, se2);
					newline(out);
					l += 2;
					break;

			case 0x5F:	l += disasmExt(out, offset); break;
			default:	if (i >= 0xE0)
					{
						// STORE_L
						se1 = pop();
						setLocal(i & 0x1F, se1);
						stack.addFirst(se1);
						printOpcode(out, String.format("STORE_L%d", i & 0x1F));
						printStack(out, se1);
						newline(out);
					}
					else if (i >= 0xC0 && i < 0xE0) 
					{
						// PUT_L
						se1 = pop();
						setLocal(i & 0x1F, se1);
						printOpcode(out, String.format("PUT_L%d", i & 0x1F));
						printStack(out, se1);
						newline(out);
					}	
					else if (i >= 0xA0 & i < 0xC0)
					{
						// PUSH_L
						se1 = getLocal(i & 0x1F);
						stack.addFirst(se1);
						printOpcode(out, String.format("PUSH_L%d", i & 0x1F));
						printStack(out, se1);
						newline(out);	
					}
					else if (i >= 60 && i < 0xA0)
					{
						int loc = i & 0x0F;
						int vec = (i & 0x30) >> 4;
						printOpcode(out, String.format("VREAD_%d_%d", loc, vec));
						newline(out);
						push(-1, "unknown");
					}
					else
					{
						printOpcode(out, String.format("?%x?", i & 0xFF));
						newline(out);
					}
		}
		if (nextLine) newline(out);
		return l;
	}


	/** Disassemble an extended opcode (one where the first byte was
	 * 0x5F).
	 * 
	 * @param out Stream to which the disassembly is being written.
	 * @param offset The address of the second byte of the opcode.
	 */
        private int disasmExt(PrintStream out, int offset)
	{
		int w1, w2;
		int l = 1;
		int i = bytecode[offset] & 0xFF;
		++offset;
		StackElement se1, se2, se3, se[];

		switch(i)
		{
			case 0x01:	se1 = pop();
					se2 = pop();
					printOpcode(out, "XOR");
					printStack  (out, se1, se2);
					newline(out);
					if (se1.value == -1 || se2.value == -1)
						push(-1, "unknown");
					else	push(se1.value ^ se2.value);
					break;
			case 0x02:	se1 = pop();
					printOpcode(out, "NOT");
					printStack  (out, se1);
					newline(out);
					if (se1.value == -1)
						push(-1, "unknown");
					else	push(~se1.value);
					break;
			case 0x03: 	se1 = pop();
					se2 = pop();
					printOpcode(out, "ROTATE");
					printStack(out, se1, se2);
					newline(out);
					push(-1, "unknown");
					break;

			case 0x04: 	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "VFIND");
					printStack(out, se1, se2, se3);
					newline(out);
					push(0, "result");
					break;	
			case 0x05: 	se1 = pop();
					se2 = pop();
					printOpcode(out, "STRCHR");
					printStack(out, se1, se2);
					newline(out);
					push(0, "result");
					break;	

			case 0x06:	se1 = pop();
				  	printOpcode(out, "PUTG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				  	printStack(out, se1);
				   	newline(out);
					++l;
				   	break;
			case 0x07:	se1 = pop();
				  	printOpcode(out, "PULLN");
				  	printStack(out, se1);
				  	printStackChain(out, se1);
				  	newline(out);
				   	break;

			case 0x09:	se1 = pop();
				  	se2 = pop();
				  	printOpcode(out, "LONGJMPR");
				  	printStack(out, se1, se2);
				  	newline(out);
				  	newline(out);
				  	break;
			case 0x0A:	se1 = pop();
				  	printOpcode(out, "LONGJMP");
				  	printStack(out, se1);
				  	newline(out);
				  	newline(out);
				  	break;
			case 0x0B:	w1 = wordFrom(offset);
				  	w2 = wordFrom(offset + 2);
				  	l += 4;
				  	printOpcode(out, "SETJMP");
				  	printArg(out, String.format("%04x,%04x", w1, w2)); 
				  	newline(out);
				 	push(0, "tuple sp");
				  	push(w1);
				  	push(w2);
				  	push(-1, "module");
				  	push(-1, "frame pointer");
				  	push(-1, "call SP");
				  	push(offset-1, "setjmp");
				  	break;
			case 0x0C: 	se1 = pop();
					se2 = pop();
					printOpcode(out, "OPEN");
					printArg(out, Integer.toString(byteFrom(offset)));
					printStack(out, String.format("(%s,%s)",
							printSE(se1),
							printSEString(se2)));
					newline(out);
					++l; 
					push(fileHandle++, "channel");
					break;
			case 0x0D: 	se1 = pop();
					printOpcode(out, "CLOSE");
					printStack(out, se1);
					newline(out);
					push(0, "result");
					break;
			case 0x0E: 	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "READ");
					printStack(out, se1, se2, se3);
					newline(out);
					push(0, "result");
					break;
			case 0x0F: 	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "WRITE");
					printStack(out, se1, se2, se3);
					newline(out);
					push(0, "result");
					break;
			case 0x10: 	se = new StackElement[4];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					printOpcode(out, "READREC");
					printStack(out, se);
					newline(out);
					push(0, "result");
					break;
			case 0x11: 	se = new StackElement[4];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					printOpcode(out, "WRITEREC");
					printStack(out, se);
					newline(out);
					push(0, "result");
					break;
			case 0x12:	printOpcode(out, "DISP");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
					++l;
				   	break;
			case 0x13:	se1 = pop();
					se2 = pop();
					printOpcode(out, "XDISP");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
					printStack(out, se1);
					printStack(out, se2);
				   	newline(out);
					++l;
				   	break;
			case 0x14: 	se1 = pop();
					printOpcode(out, "FSIZE");
					printStack(out, se1);
					newline(out);
					push(-1, "filesize");
					break;
			case 0x15: 	se1 = pop();
					printOpcode(out, "UNLINK");
					printStack(out, String.format("(%s)",
							printSEString(se1)));
					newline(out);
					++l; 
					push(0, "result");
					break;

			case 0x16:	printOpcode(out, "PULLRET");
					newline(out);
					break;
			case 0x17:	printOpcode(out, "KBINPUT");
					newline(out);
					push(0, "key");
					break;
			case 0x1C:	se1 = pop();
					printOpcode(out, "TPULL");
					printStack(out, se1);
					newline(out);
					break;
			case 0x1F: 	se = new StackElement[5];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					se[4] = pop();
					printOpcode(out, "STRICMP");
					printStack(out, se);
					newline(out);
					push(0, "result");
					break;
			case 0x20: 	se = new StackElement[5];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					se[4] = pop();
					printOpcode(out, "STRICMP1");
					printStack(out, se);
					newline(out);
					push(0, "result");
					break;

			case 0x24:	se1 = pop();
					se2 = pop();
					se3 = pop();
					printOpcode(out, "MEMCMP");
					printStack(out, se1, se2, se3);
					newline(out);
					push(0, "result");
					break;
			case 0x25: 	se = new StackElement[4];
					se[0] = pop();
					se[1] = pop();
					se[2] = pop();
					se[3] = pop();
					printOpcode(out, "MEMCMPO");
					printStack(out, se);
					newline(out);
					push(0, "result");
					break;
			case 0x26:	se1 = pop();
					se2 = pop();
					printOpcode(out, "VECL");
					printStack(out, se1, se2);
					newline(out);
					push(0, "result");
					break;

			case 0x27:	printOpcode(out, "DECMG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
					push(byteFrom(offset), "sysvar");
					++l;
				   	break;
			case 0x28:	printOpcode(out, "DECG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
					push(byteFrom(offset), "global");
					++l;
				   	break;
			case 0x29:	printOpcode(out, "PULL_");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
					se = new StackElement[byteFrom(offset)];
					for (int n = 0; n < se.length; n++)
					{
						se[n] = pop();
					}
					printStack(out, se);
					newline(out);
					++l;
					break;	
			case 0x2A:	printOpcode(out, "INCG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
					++l;
					break;
			case 0x2B:	printOpcode(out, "INCMG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
					++l;
				   	break;
			case 0x2C:	se1 = pop();
					push(se1.value, se1.type);
				  	printOpcode(out, "STOREMG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				  	printStack(out, se1);
				   	newline(out);
					++l;
				   	break;
			case 0x2D:	se1 = pop();
					push(se1.value, se1.type);
				  	printOpcode(out, "STOREG");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				  	printStack(out, se1);
				   	newline(out);
					++l;
					break;
			case 0x2E:	printOpcode(out, "RET_");
				  	printArg(out, String.format("%d", 
						byteFrom(offset)));
				   	newline(out);
				   	newline(out);
					++l;
					break;
			case 0x2F:	se1 = pop();
					printOpcode(out, "PRCHAR");
					printStack(out, se1);
					newline(out);
					break;
			case 0x38:	se1 = pop();
					se2 = pop();
					printOpcode(out, "RENAME");
					printStack(out, se1);
					newline(out);
					push(0, "result");
					break;
			default:
				printOpcode(out, String.format("EXT:%02X", i));
				newline(out);
				break;
		}
		return l;

	}



	/** Get the number of bytes in an opcode.
	 *
	 * Note: disasmInst() doesn't use this; it has its own copy of the
	 * opcode-length code. Really it should use this instead.
	 *
	 * @param offset The offset of the start of the opcode.
	 * @return The number of bytes to the next opcode. Bit 15 is set if
	 * 	  execution cannot flow past this opcode (JUMP, HALT, a 
	 *	  RETURN-type opcode, etc.) 
	 */
	private int opcodeLength(int offset)
	{
		int l = 1;
		int i = bytecode[offset] & 0xFF;

		++offset;
		switch(i)
		{
			case 0x00:	l |= 0x8000; break;	// BREAK
			case 0x08:				// INCL
			case 0x0B:				// DECL
			case 0x18:				// VLOADW_
			case 0x19:				// VLOADB_
			case 0x1C:				// VPUTW_
			case 0x1D:				// VPUTW_
			case 0x22:				// LOADG
			case 0x23:	l++; break;		// LOADMG
			case 0x25:	l += 2; break;		// PUSHW
			case 0x26:				// PUSHB
			case 0x2D:	l++; break;		// PUTMG
			case 0x30:	l += jumpLength(offset);// JUMP
					l |= 0x8000;
					break;
			case 0x31:				
			case 0x32:	
			case 0x33:	
			case 0x34:	
			case 0x35:	
			case 0x36:	
			case 0x37:	
			case 0x38:	
			case 0x39:	
			case 0x3A:	
			case 0x3B:	
			case 0x3C:	
			case 0x3D:	
			case 0x3E:	l += jumpLength(offset); break; //JUMPc
			case 0x3F:
			case 0x40:
			case 0x41:
			case 0x42:	l += 2; break;		// CALLn
			case 0x43:	l += 1; break;		// CALL
			case 0x44:	
			case 0x45:	
			case 0x46:
			case 0x47:	l += 2; break;		// CALLFn
			case 0x48:	l += 1; break;		// CALLF
			case 0x49:				// RETURN
			case 0x4A:				// RFALSE
			case 0x4B:	l |= 0x8000; break;	// RZERO
			case 0x4D:	l += 2;			// HALT
					/*l |= 0x8000; */break;
			case 0x4E:	l += nextb(offset);
					break;			// NEXTB
			case 0x54:	l++; break;		// INCLV
			case 0x55:	l |= 0x8000; break;	// RET
			case 0x56:	l++; break;		// PUTL
			case 0x57:	l++; break;		// LOADL
			case 0x58:	l++; break;		// STOREL
			case 0x59:				// BITSVL
			case 0x5A:				// BITSV
			case 0x5B:				// BBSETVL
			case 0x5C:				// BBSETV
			case 0x5D:				// BSETVL
			case 0x5E:	l += 2; break;		// BSETV
			case 0x5F:	l += extLength(offset); 
					break;
		}
		return l;
	}

	/** Get the number of bytes in a jump opcode.
	 *
	 * @param offset The offset of the second byte of the opcode.
	 * @return The number of bytes to the next opcode.*/
	private int jumpLength(int offset)
	{
		if (bytecode[offset] == 0)
		{
			return 3;
		}
		else
		{
			return 1;
		} 
	}

	/** Get the number of bytes in an extended opcode.
	 *
	 * @param offset The offset of the second byte of the opcode.
	 * @return The number of bytes to the next opcode.*/
        private int extLength(int offset)
	{
		int l = 1;
		int i = bytecode[offset] & 0xFF;
		++offset;

		switch(i)
		{
			case 0x06:	++l; break;	// PUTG
			case 0x0B:	l += 4; break;	// SETJMP
			case 0x0C: 	++l; break;	// OPEN
			case 0x12:	++l; break;	// DISP
			case 0x13:	++l; break;	// XDISP
			case 0x27: 	++l; break;	// DECMG
			case 0x28: 	++l; break;	// DECG
			case 0x29: 	++l; break;	// PULL_
			case 0x2A: 	++l; break;	// INCG
			case 0x2B: 	++l; break;	// INCMG
			case 0x2C: 	++l; break;	// STOREMG
			case 0x2D: 	++l; break;	// STOREG
			case 0x2E: 	++l; 
					l |= 0x8000; break;	// RET_

		}
		return l;

	}

	/** Analyse a module and work out where all its procedures and 
	 * jump targets are.
	 *
	 * @param symbols The symbol table, which will be populated with
	 *  a list of processes and labels.
	 */
	public void analyseModule(SymbolTable symbols)
	{
		int n;
		int offset;

		// Initialise hitMap to all zeroes.
		hitMap = new byte[bytecode.length];	
		for (n = 0; n < hitMap.length; n++)
		{
			hitMap[n] = 0;
		}
		// Analyse all public procedures.
		for (n = 0; n < header.getSize(); n++)
		{
			analyseProc(header.getProcAddress(n), symbols);
		}
		// That may have found more procedures; do them as well.
		while (true)
		{
			// Search for non-analysed procedures.
			Symbol[] syms = symbols.moduleSymbols(moduleID + 1, 'p');
			if (syms.length == 0) break;	
			// And analyse each one.
			for (Symbol s: syms)
			{
				analyseProc(s.offset, symbols);
			}
		}
	}


	/** Analyse a procedure.
	 * @param offset The address of the procedure header.
	 * @param symbols The symbol table in the process of being populated.
	 */
	public void analyseProc(int offset, SymbolTable symbols)
	{
		/* Add a symbol for this procedure and give it type P 
		 * (Procedure, analysed) */
		symbols.add('P', String.format("P%02d%04X", moduleID + 1, 
			offset), moduleID + 1, offset);

		if (offset >= bytecode.length) 
		{
			System.err.printf("Procedure %04x out of range for module %d\n", offset, moduleID + 1);
			return;
		}
		int base = offset;
		int procHead = bytecode[offset] & 0xFF;
		SymbolTable localJumps = new SymbolTable();

		// Skip over variable initialisation.
		if ((procHead & 0x80) != 0)
		{
			byte bt = 0;
			short val;
			while ((bt & 0x80) == 0)
			{
				bt = bytecode[++offset];

				if ((bt & 0x40) != 0) 
				{
					++offset;
				}
				else
				{
					offset += 2;
				}
			}	

		}
		++offset;
		// Analyse the first chunk of code, up to the first 
		// break in the flow of control (HALT, JUMP, RET etc.)
		int max = analyseBlock(offset, symbols, localJumps, -1);
		int min = base;

		// The chunk we just analysed may have included jumps to 
		// nearby bits of code. Look in localJumps and see if 
		// there's one immediately adjacent to the block we've
		// just done.

		boolean anotherBlock;
		do
		{
			anotherBlock = false;
			Symbol s = localJumps.lookup(0, max);
			if (s != null && s.type == 'l')
			{
				// Mark jump entry as processed
				s.type = 'L';	
				anotherBlock = true;	
				// And find new upper limit of analysed code.
				max = analyseBlock(max, symbols, localJumps, -1);
				break;	
			}
		} while (anotherBlock);

		// There may also be jumps to blocks of code that aren't
		// contiguous with the bit we've just analysed.
		// Analyse them as well.
		do
		{
			// Get list of pending local labels.
			Symbol [] syms = localJumps.moduleSymbols(0, 'l');
			anotherBlock = false;
			for (Symbol s : syms)
			{
				if (s.type == 'l')
				{
					// Mark label as processed.
					s.type = 'L';
					anotherBlock = true;
// If it's before the current block, go up to the start of the current block.
					if (s.offset < min)
					{
						analyseBlock(s.offset, 
						 symbols, localJumps, min);
						min = s.offset;
					}
					else
					{	
// If it's not, it must be in between or after. Analyse it, and if that
// goes further than we've analysed so far, increase 'max'.
						int altmax = analyseBlock(
						 s.offset, symbols, localJumps,
						 -1);
						if (altmax > max) max = altmax;
					}
					break;
				}
			}
		} while (anotherBlock);
		// Set the length of the procedure.
		Symbol ps = symbols.lookup(moduleID + 1, base);
		ps.max = max;
	}

	/** Analyse a block of code.
	 *
	 * @param offset The first instruction to analyse.
	 * @param symbols The global symbol table.
	 * @param localJumps The list of local jumps in the current procedure.
	 * @param stopAt Stop analysing once this address has been reached.
	 * @return The first address not analysed.
	 */
	private int analyseBlock(int offset, SymbolTable symbols, 
					SymbolTable localJumps, int stopAt)
	{
		int l = 0;
		boolean stopped = false;
		boolean noreturn = false;
		int lastPush = -1;
		byte lastOpcode = 0x00;

		while (!stopped)	
		{
			Symbol sym;
			if (offset >= bytecode.length) break;
//			System.err.printf("Analyse %02d:%04x\n",
//				moduleID + 1, offset);
			sym = localJumps.lookup(0, offset);

// Mark any local labels we come across as visited.
			if (sym != null && sym.type == 'l')
			{
				sym.type = 'L';
			}
			hitMap[offset]++;
			if (hitMap[offset] > 4)
			{
				System.err.println("Examining the same bit of code twice\n");
				return offset;
			}
			switch (bytecode[offset] & 0xFF)
			{
				case 0x30:
				case 0x31:
				case 0x32:
				case 0x33:
				case 0x34:
				case 0x35:
				case 0x36:
				case 0x37:
				case 0x38:
				case 0x39:
				case 0x3A:
				case 0x3B:
				case 0x3C:
				case 0x3D:
				case 0x3E:
					analyseJump(offset + 1, symbols, localJumps);
					break;

				case 0x3F:
				case 0x40:
				case 0x41:
				case 0x42: /* Local CALL */
					noreturn = analyseCall(wordFrom(offset + 1), symbols);
					break;
				case 0x43:
					/* Computed CALL */
//					System.err.printf("lastPush=%04x\n", lastPush);
					if (lastPush != -1)
						noreturn = analyseCall(lastPush, symbols);
					break;
				case 0x44:
				case 0x45:
				case 0x46:
				case 0x47:
					noreturn = analyseFarCall(wordFrom(offset + 1), symbols);
					break;
				case 0x48:
					/* Computed FARCALL */
//					System.err.printf("lastPush=%04x\n", lastPush);
					if (lastPush != -1)
						noreturn = analyseFarCall(lastPush, symbols);
					break;

				default:
					break;
			}
			lastOpcode = bytecode[offset];
			if (lastOpcode == 0x25)
				lastPush = wordFrom(offset+1); 
			else	lastPush = -1;
			l = opcodeLength(offset);
			if (noreturn) l |= 0x8000;
			offset += (l & 0x7FFF);
			if ((l & 0x8000) != 0) stopped = true;
			if (stopAt >= 0 && offset >= stopAt) break;
		}
		return offset;
	}

	/** Analyse a CALL (near) opcode.
	 *
	 * @param target The address of the target procedure.
	 * @param symbols The global symbol table.
	 * @return true if the procedure does not return, else false.
	 */
	private boolean analyseCall(int target, SymbolTable symbols)
	{
		Symbol sym;
		boolean noreturn = false;

		// See if it's marked as 'does not return'.	
		sym = parent.controls.lookup(moduleID + 1, target);
		if (sym != null && sym.type == 'N')
		{
			noreturn = true;
		}
		// See if this procedure is known about.
		sym = symbols.lookup(moduleID + 1, target);

		if (sym == null)
		{
			// No, not previously encountered. Add it with type
			// 'p' (needs analysing).
			symbols.add('p', String.format("P%02d%04X", 
				moduleID + 1, target), moduleID + 1, target);
		}
		return noreturn;
	}

	/** Analyse a CALL (far) opcode.
	 *
	 * @param target The address of the target procedure.
	 * @param symbols The global symbol table.
	 * @return true if the procedure does not return, else false.
	 */
	private boolean analyseFarCall(int target, SymbolTable symbols)
	{
		int proc = target & 0xFF;	 // Procedure ID in module
		int mod  = (target >> 8) & 0xFF; // 1-based module ID 
		Symbol sym;
		boolean noreturn = false;
	
		// Check module ID is in range (if it isn't there's 
		// probably a fault in the analyser).	
		if (mod <= 0 || mod > parent.modheads.length)
		{
			return false;
		}
		// Get the offset of the procedure in the other module.
		int xoffset = parent.modheads[mod - 1].getProcAddress(proc);
	
		// See if it's marked as 'does not return'.	
		sym = parent.controls.lookup(mod, xoffset);
		if (sym != null && sym.type == 'N')
		{
			noreturn = true;
		}
		// See if this procedure is known about.
		sym = symbols.lookup(mod, xoffset);

		if (sym == null)
		{
			// No, not previously encountered. Add it with type
			// 'p' (needs analysing).
			symbols.add('p', String.format("P%02d%04X", mod, 
				xoffset), mod, xoffset);
		}
		return noreturn;
	}


	/** Analyse a JUMP opcode.
	 * @param offset The offset of the byte after the jump instruction.
	 * @param symbols The symbol table in which persistent labels are 
	 * 		  recorded.
	 * @param localJumps The symbol table in which jumps in the current
	 * 		  procedure are recorded.
	 */
	private void analyseJump(int offset, 
			SymbolTable symbols,
			SymbolTable localJumps)
	{
		int target;

		if (bytecode[offset] == 0)
		{
			// Absolute jump 
			target = wordFrom(offset + 1);
		}
		else
		{
			// Relative jump
			int b = bytecode[offset];
	
			if (b < 0) target = offset + 1 + b;
			else	   target = offset + 1 + b;
		}
		// Is this already a known label?
		Symbol sym = symbols.lookup(moduleID, target);

		// If not, add it with type 'L'.
		if (sym == null)
		{
			symbols.add('L', String.format("L%02d%04X", 
				moduleID + 1, target), moduleID + 1, target);
		}
		// Does it point to an area of code that hasn't yet been
		// analysed? If so, add it to localJumps with type 'l'.
		sym = localJumps.lookup(moduleID, target);
		if (sym == null && hitMap[target] == 0) 
		{
			localJumps.add('l', "", 0, target);
		}
	}


	/** Display code and data areas for this module.
	 *
	 * Does not amalgamate adjacent code areas, which perhaps it should.
	 *
	 * @param out The stream to which the output should be written.
	 * @param symbols The symbol table used to calculate procedure 
	 * lengths.
	 */
        public void coverage(PrintStream out, SymbolTable symbols)
	{
		int n;
		int offset=0, pos = 0;
		List<Integer> procList = new ArrayList<Integer>();

		// Get a list of all procedures in this module.
		Symbol sym[] = symbols.moduleSymbols(moduleID + 1, 'P');
		for (Symbol s : sym)
		{
//			System.err.printf(">>%s %c:%02d:%04x\n", 
//				s.name, s.type, s.module, s.offset);
			procList.add(s.offset);
		}
		// Sort it.
		Collections.sort(procList);
		// Then for each procedure:
		for (n = 0; n < procList.size(); n++)
		{
			// Get its start and end.
			offset = procList.get(n);
			Symbol s = symbols.lookup(moduleID + 1, offset);
			// If there's a gap between the last procedure and
			// this one, report it.
			if (offset > pos)
			{
				// Is this an orphaned jump?
				if (bytecode[pos] == 0x30 && 
					((offset - pos) == 2 || 
					 (offset - pos) == 4))
				{
					out.printf("; %02d:%04x - %02d:%04x  jump?\n",
					moduleID + 1, pos, moduleID + 1, offset);
	
				}
				else	// No. Report it as data.
				{
					out.printf("; %02d:%04x - %02d:%04x  data\n",
					moduleID + 1, pos, moduleID + 1, offset);
				}
			}
			// If this procedure starts before the end of the
			// previous procedure, we have an overlap (probably
			// a fault in the analyser or disassembler).
			if (offset < pos)
			{
				out.printf(";! Overlapping procedures %04x vs %04x\n", 
					pos, offset);

			}
			// Print range of this procedure 
			out.printf("; %02d:%04x - %02d:%04x  code\n",
					moduleID + 1, offset, 
					moduleID + 1, s.max);
			pos = s.max;
		}
		// Check for data after the last procedure.
		if (pos < bytecode.length)
		{
				out.printf("; %02d:%04x - %02d:%04x  data\n",
				moduleID + 1, pos, moduleID + 1, bytecode.length);
		}
	}

	/** Dump a data area. 
	 * @param out The stream to which the data area should be 
	 * written.
	 * @param from The first address in the data area.
	 * @param upto The first address beyond the data area. */
	void dumpData(PrintStream out, int from, int upto)
	{
		int n;

		// If the data area is 2 or 4 bytes long, and the first 
		// byte is 0x30, then it's probably an unreachable 
		// jump. Whatever compiled Cornerstone seems rather inclined 
		// to include these.
		if ((upto == from + 2 || upto == from + 4)  &&
			bytecode[from] == 0x30)
		{
			out.printf("%02d:%04x    [%02x]               ",
				moduleID + 1, from, bytecode[from]);
			guessJump(out, from);
			newline(out);
			return;
		}

		// Otherwise, just spit out the data, 1 byte / line.
		for (n = from; n < upto; n++)
		{
			out.printf("%02d:%04x    [%02x]               DEFB       ",
				moduleID + 1, n, bytecode[n]);
			if (bytecode[n] >= 0x20 && bytecode[n] < 0x7F)
			{
				out.printf(" '%c'\n", bytecode[n]);	
			}
			else	out.printf(" %d\n", bytecode[n] & 0xFF);
		}
		newline(out);
				
	}

}





