/* CSD: Cornerstone disassembler

    Copyright (C) 2014  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

package info.seasip.csd;

import java.io.*;

/** Represents the module header for a given module. The module header 
 * contains a list of entry points.
 *
 * @author      John Elliott <jce@seasip.demon.co.uk>
 * @version     1.0
 * @since       2014-04-12
 */
public class ModuleHeader extends WordChunk
{
	/** The number of entry points defined by this header. */
	private int size;
	/** The number of this module, 0-based. */
	private int module;
	/** The offset of this particular record in the .MME file. */
	private long base;

	/** Standard constructor.
	 * @param rf The .MME file.
	 * @param m  The module ID, 0-based.
	 */
	ModuleHeader(RandomAccessFile rf, int m)
	{
		super(rf);
		module = m;
	}

	/** Load the module header.
	 * @throws IOException on read error. */
	public void load() throws IOException
	{
		base = file.getFilePointer();
		size = read();	// Number of words that follow.
		read(size);	// Read them.
	}

	/** Return the number of entry points.
	 * @return The number of entry points. */
	public int getSize() 
	{ 
		return size;
	}


	/** Convert an entry point index to an offset.
	 * @param sym The index of the entry point.
	 * @return If sym is a valid index, the offset of that entry point;	
	 *        otherwise, 0. */
	public int getProcAddress(int sym)
	{
		if (sym < 0 || sym >= words.length) return 0;
		return words[sym];
	}

	/** Dump the contents of the module header for this module.
	 * @param stream The stream to output to.
	 */
	public void showContents(PrintStream stream)
	{
		long pos;

		stream.println(";---------------------------------------");
		stream.printf ("; Module %d header:\n", module + 1);
		stream.println(";---------------------------------------");
		stream.printf ("; %04x: M%d header length = %d words\n",
				base, module + 1, size);
		for (int n = 0; n < size; n++)
		{
			stream.printf ("; %04x: Proc %d at %04x", 
				base + 2 * n + 2, n, words[n]);

			stream.print("\n");
		}
	}
	
}

